use moksha::types::Result;
use moksha::errors::Error;
use moksha::state::{AbstractState, GlobalState};
use moksha::state::fs_store::*;
use moksha::fs::fs_factory::*;
use moksha::fs::btrfs_desc::BtrfsDesc;
use moksha::fs::targz_fs_desc::TargzFsDesc;
use num_traits::ToPrimitive;

use std::path::PathBuf;

use dkregistry::v2::Client;

// use tokio::prelude::*;
// use futures::executor::block_on;
use futures::*;
use std::sync::Arc;

use clap::{App, Arg};

use log;

const DESCRIPTION: &str = "moksha-state helps you manage your moksha state store";
const DEFAULT_STORE_ROOT: &str = "/var/lib/moksha";

enum Command {
    Init,
    Pull,
    NotImplemented
}

impl From<&str> for Command {
    fn from(str_command: &str) -> Command {
        match str_command {
            "init" => Command::Init,
            "pull" => Command::Pull,
            _ => Command::NotImplemented,
        }
    }
}

// async fn main() -> Result<(), Box<dyn std::error::Error>> {
#[tokio::main]
async fn main() -> Result<()> {
    env_logger::Builder::from_env(env_logger::Env::default().default_filter_or("info")).init();

    let arg_matches = App::new(env!("CARGO_PKG_NAME"))
        .version(env!("CARGO_PKG_VERSION"))
        .author(env!("CARGO_PKG_AUTHORS"))
        .about(DESCRIPTION)
        .subcommand(
            App::new("init")
                .arg(
                    Arg::with_name("root_dir")
                        .short("b")
                        .long("root-dir")
                        .takes_value(true)
                        .help("root dir (default: '/var/lib/moksha')"),
                )
        )
        .subcommand(
            App::new("pull")
                .arg(
                    Arg::with_name("root_dir")
                        .short("b")
                        .long("root-dir")
                        .takes_value(true)
                        .help("root dir (default: '/var/lib/moksha')"),
                )
                .arg(
                    Arg::with_name("image")
                        .short("i")
                        .long("image")
                        .takes_value(true)
                        .help("image to pull")
                )
        )
        .get_matches();

    let command_tuple = arg_matches.subcommand();
    let command: Command = command_tuple.0.into();
    let command_args = command_tuple.1;

    log::debug!("moksha-store started");

    match (command, command_args) {
        (Command::Init, Some(list_opts)) => {
            let store_root = list_opts
                .value_of("root_dir")
                .unwrap_or(DEFAULT_STORE_ROOT)
                .to_string();

            let store_root_path = PathBuf::from(store_root);
            
            let mut global_state = GlobalState::open(
                store_root_path,
                Some(GlobalState::empty_variant()))?;

            let test_item = &mut FsStoreItem {
                item_name: FsStoreItemName {
                    domain: "kitzplatz".to_string(),
                    name: "dummy".to_string(),
                    version: "0.1".to_string(),
                    overlay: 0
                },
                item_hash: FsStoreItemId {
                    id: "".to_string()
                },
                prev_overlay: None,
                fs_type: FsItemType::Btrfs,
                encryption: None,
                contents_checksum: None,
                base_path: PathBuf::new()
            };
            
            global_state.fs_store.add_item(test_item)?;
            
            Ok(())
        }
        (Command::Pull, Some(list_opts)) => {
            let store_root = list_opts
                .value_of("root_dir")
                .unwrap_or(DEFAULT_STORE_ROOT)
                .to_string();

            let store_root_path = PathBuf::from(store_root);

            let image_id_opt = list_opts
                .value_of("image");

            if image_id_opt.is_none() {
                log::error!("command {} requires an 'image' argument", command_tuple.0);
                return Ok(println!("{}", arg_matches.usage()));
            }

            let image_id = image_id_opt.unwrap().split_terminator(":").collect::<Vec<&str>>();
            let image_name = *image_id.first().unwrap();
            let image_version = *image_id.last().unwrap();
            
            let mut global_state = GlobalState::open(
                store_root_path,
                None)?;

            let registry_client = Client::configure()
                .registry("registry-1.docker.io")
                .build()?
                .authenticate(&[format!("repository:{}:pull", image_name).as_str()]).await?;

            registry_client.is_auth().await?;

            let image_manifest = registry_client
                .get_manifest(image_name, image_version)
                .await?;

            // log::info!("manifest: {:?}", image_manifest);
            log::info!("successfully pulled manifest for {}:{}", image_name, image_version);

            // let mut last_fs_item = None;

            log::info!("pulling layers...");

            let blob_results: Vec<Result<Vec<u8>>> = future::join_all(
                image_manifest
                    .layers_digests(None)?
                    .iter()
                    .map(Arc::new)
                    .map(|layer| {
                        log::info!("pulling {}", layer);
                        
                        registry_client
                            .get_blob(image_name, layer.as_str())
                            .and_then(|blob| async move {
                                log::info!("finished pulling {}", layer);
                                Ok(blob)
                            })
                            .map_err(Error::DockerRegistryError)
                    }))
                .await;

            log::info!("finished pulling");
            log::info!("writing to disk");
            
            let mut last_overlay = None;

            // processing blobs
            // todo: write sha256 as an attr, check for already existing images
            // make this part of a pulling fiber
            for (i, blob_result) in blob_results.iter().enumerate() {
                let blob = blob_result.as_deref().map_err(|e| {
                    log::error!("unable to pull blob: {}", e);
                    Error::StateError(format!("unable to pull blob: {}", e))
                })?;
                
                let new_item = &mut FsStoreItem {
                    item_name: FsStoreItemName {
                        domain: "docker".to_string(),
                        name: image_name.to_string(),
                        version: image_version.to_string(),
                        overlay: i.to_u64().unwrap()
                    },
                    item_hash: FsStoreItemId {
                        id: "".to_string()
                    },
                    prev_overlay: last_overlay.clone(),
                    fs_type: FsItemType::Btrfs,
                    encryption: None,
                    contents_checksum: None,
                    base_path: PathBuf::new()
                };

                // adding of new item, and data backup the tar
                // todo check for layer type
                new_item.reset_id();
                log::debug!("processing blob {}", new_item.item_hash.id.clone());
                
                global_state.fs_store.add_item(new_item)?;
                new_item.write_attr_raw("backup.tar.gz".into(), blob.to_vec())?;

                log::debug!("wrote backup for {}", new_item.item_hash.id.clone());
                log::debug!("writing btrfs fs for {}", new_item.item_hash.id.clone());
                
                new_item.wait_for_lock()?;
                new_item.lock()?;
                let mut fs_factory =
                    FsFactory::<TargzFsDesc, BtrfsDesc>::new(new_item.base_path().join("image"));

                fs_factory.set_data(blob.to_vec())
                    .or_else(|e| { new_item.unlock()?; Err(e) })?;
                fs_factory.finish()
                    .or_else(|e| { new_item.unlock()?; Err(e) })?;
                
                new_item.unlock()?;
                
                last_overlay = Some(new_item.item_hash.id.clone());

                log::debug!("done processing {}", new_item.item_hash.id.clone());
            }

            log::info!("finished persisting");
            
            return Ok(())
        }
        (_, _) => {
            log::error!("command {} not implemented", command_tuple.0);
            Ok(println!("{}", arg_matches.usage()))
        }
    }
}
