use moksha::unix::ProcHandler;

use moksha::namespace::{NamespaceBundle, NamespaceBundleStruct, NamespaceOption, NamespaceType};
use moksha::provider::{JailProvider, NamespaceProvider};
use moksha::types::{Pid, Result};

use moksha::errors::Error;

use clap::{App, Arg};

use std::path::PathBuf;
use std::str::FromStr;

use log;

static DESCRIPTION: &str = "nstool is an utility for managing linux namespaces";

enum Command {
    List,
    Exec,
    Enter,
    NotImplemented,
}

impl From<&str> for Command {
    fn from(str_command: &str) -> Command {
        match str_command {
            "ls" => Command::List,
            "exec" => Command::Exec,
            "enter" => Command::Enter,
            _ => Command::NotImplemented,
        }
    }
}

fn main() -> Result<()> {
    env_logger::Builder::from_env(env_logger::Env::default().default_filter_or("info")).init();

    let arg_matches = App::new(env!("CARGO_PKG_NAME"))
        .version(env!("CARGO_PKG_VERSION"))
        .author(env!("CARGO_PKG_AUTHORS"))
        .about(DESCRIPTION)
        .subcommand(
            App::new("ls")
                .arg(
                    Arg::with_name("root_dir")
                        .short("b")
                        .long("root-dir")
                        .takes_value(true)
                        .help("root dir (default: '/')"),
                )
                .arg(
                    Arg::with_name("interesting_namespaces")
                        .long("interesting-namespaces")
                        .short("i")
                        .takes_value(true)
                        .help("interesting namespaces"),
                ),
        )
        .subcommand(
            App::new("exec")
                .arg(
                    Arg::with_name("root_dir")
                        .short("b")
                        .long("root-dir")
                        .takes_value(true)
                        .help("root dir (default: '/')"),
                )
                .arg(
                    Arg::with_name("new_root")
                        .short("r")
                        .long("new-root")
                        .takes_value(true)
                        .help("root dir"),
                )
                .arg(
                    Arg::with_name("namespaces")
                        .long("namespaces")
                        .short("n")
                        .takes_value(true)
                        .help("namespaces to be entered"),
                )
                .arg(
                    Arg::with_name("command")
                        .long("command")
                        .short("c")
                        .required(true)
                        .takes_value(true)
                        .multiple(true),
                ),
        )
        .subcommand(
            App::new("enter")
                .arg(
                    Arg::with_name("root_dir")
                        .short("b")
                        .long("root-dir")
                        .takes_value(true)
                        .help("root dir (default: '/')"),
                )
                .arg(
                    Arg::with_name("pid")
                        .short("p")
                        .long("pid")
                        .required(true)
                        .takes_value(true)
                        .help("pid of which namespaces should be entered"),
                )
                .arg(
                    Arg::with_name("namespaces")
                        .long("namespaces")
                        .short("n")
                        .takes_value(true)
                        .help("namespaces to be entered"),
                )
                .arg(
                    Arg::with_name("command")
                        .long("command")
                        .short("c")
                        .required(true)
                        .takes_value(true)
                        .multiple(true),
                ),
        )
        .get_matches();

    let command_tuple = arg_matches.subcommand();
    let command: Command = command_tuple.0.into();
    let command_args = command_tuple.1;

    log::debug!("nstool started");

    match (command, command_args) {
        (Command::List, Some(list_opts)) => {
            let root_dir = PathBuf::from_str(list_opts.value_of("root_dir").unwrap_or("/"))?;
            let interesting_namespaces = list_opts
                .value_of("interesting_namespaces")
                .map(|ns_list| {
                    ns_list
                        .split(",")
                        .into_iter()
                        .map(|ine| {
                            NamespaceOption::from_str(ine)
                                .map(Some)
                                .unwrap_or_else(|e| {
                                    log::warn!("unable to parse {}, error: {}", ine, e);
                                    None
                                })
                        })
                        .flatten()
                        .collect::<Vec<NamespaceOption>>()
                })
                .unwrap_or(vec![
                    NamespaceOption::Pid,
                    NamespaceOption::Mnt,
                    NamespaceOption::Net,
                ]);

            let handler = ProcHandler::new(root_dir);
            let pids = &handler.get_pids()?;

            let pid_processor = |pid: &Pid| -> Result<NamespaceBundle> {
                interesting_namespaces
                    .iter()
                    .map(|&nso| Ok((pid.clone(), nso, handler.get_namespace(pid.clone(), nso)?)))
                    .collect::<Result<NamespaceBundle>>()
            };

            let error_handler = |r: Result<NamespaceBundle>| {
                r.map(Some).or_else(|e: Error| match e {
                    Error::PidNotFound(_) => {
                        log::warn!("non-fatal: {}", e);
                        Ok(None)
                    }
                    _ => Err(e),
                })
            };

            let entries = pids
                .into_iter()
                .map(pid_processor)
                .map(error_handler)
                .filter(|r| match r {
                    Ok(o) => match o {
                        Some(_) => true,
                        None => false,
                    },
                    Err(_) => true,
                })
                .map(|r| r.map(|o| o.unwrap()))
                .collect::<Result<Vec<NamespaceBundle>>>()
                .map(|bundles| bundles.into_iter().flatten().collect::<NamespaceBundle>())?;

            let ls_output = serde_json::to_string_pretty(&NamespaceBundleStruct { 0: entries })?;

            println!("{}", ls_output);

            log::debug!("nsstat finished");
            Ok(())
        }
        (Command::Exec, Some(list_opts)) => {
            let root_dir = PathBuf::from_str(list_opts.value_of("root_dir").unwrap_or("/"))?;
            let new_root = list_opts
                .value_of("new_root")
                .map_or(Ok(None), |nr| PathBuf::from_str(nr).map(Some))?;

            let command_argv = list_opts
                .values_of("command")
                .unwrap()
                .map(|s| s.to_string())
                .collect::<Vec<String>>();

            if command_argv.len() == 0 {
                return Err(Error::CLIError("empty command supplied".to_string()));
            }

            let namespaces = list_opts
                .value_of("namespaces")
                .map(|ns_list| {
                    ns_list
                        .split(",")
                        .into_iter()
                        .map(|ine| {
                            NamespaceType::from_str(ine).map(Some).unwrap_or_else(|e| {
                                log::warn!("unable to parse {}, error: {}", ine, e);
                                None
                            })
                        })
                        .flatten()
                        .collect::<Vec<NamespaceType>>()
                })
                .unwrap_or(vec![
                    NamespaceType::Pid,
                    NamespaceType::Mnt,
                    NamespaceType::Net,
                ]);

            let jail_provider: Box<dyn JailProvider> = Box::new(NamespaceProvider::new(
                root_dir,
                new_root,
                None,
                namespaces,
                PathBuf::from_str(command_argv.first().unwrap())?,
                command_argv,
            ));

            jail_provider.exec()
        }
        (Command::Enter, Some(list_opts)) => {
            let root_dir = PathBuf::from_str(list_opts.value_of("root_dir").unwrap_or("/"))?;
            let pid = list_opts.value_of("pid").unwrap().parse()?;

            let command_argv = list_opts
                .values_of("command")
                .unwrap()
                .map(|s| s.to_string())
                .collect::<Vec<String>>();

            if command_argv.len() == 0 {
                return Err(Error::CLIError("empty command supplied".to_string()));
            }

            let namespaces = list_opts
                .value_of("namespaces")
                .map(|ns_list| {
                    ns_list
                        .split(",")
                        .into_iter()
                        .map(|ine| {
                            NamespaceType::from_str(ine).map(Some).unwrap_or_else(|e| {
                                log::warn!("unable to parse {}, error: {}", ine, e);
                                None
                            })
                        })
                        .flatten()
                        .collect::<Vec<NamespaceType>>()
                })
                .unwrap_or(vec![
                    NamespaceType::Pid,
                    NamespaceType::Mnt,
                    NamespaceType::Net,
                ]);

            let jail_provider: Box<dyn JailProvider> = Box::new(NamespaceProvider::new(
                root_dir,
                None,
                Some(pid),
                namespaces,
                PathBuf::from_str(command_argv.first().unwrap())?,
                command_argv,
            ));

            jail_provider.exec()
        }
        _ => Ok(println!("{}", arg_matches.usage())),
    }
}
