use std::{error, fmt};

use clap::Error as ClapError;
use regex::Error as RegexError;
// use fs_extra::error::{Error as FsError};
    
use std::convert::Infallible as StdInfallible;

use serde_json::Error as SerdeJsonError;

use std::ffi::NulError;

use std::cell::{BorrowError, BorrowMutError};
use std::io::Error as IOError;
use std::num::ParseIntError;
use std::path::StripPrefixError;

use dkregistry::errors::{Error as DkRegError};

use crate::types::Pid;

#[derive(Debug)]
pub enum Error {
    CLIError(String),

    ProviderError(String),

    NamespaceLinkNameError(String),
    NamespaceParsingError(String),
    NamespaceValidationError(String),

    LoopCtrlError(String),
    LoopDeviceError(String),

    StateError(String),
    StripPrefixError(StripPrefixError),
    StateGetItemError(String),

    FilesystemFactoryError(String),
    FilesystemDescError(String),
    
    DockerRegistryError(DkRegError),
    DockerImageSourceError(String),

    WriterParseError(String),
    Infallible(StdInfallible),    
    CStringError(NulError),

    SerdeJsonError(SerdeJsonError),

    ClapError(ClapError),

    BorrowError(BorrowError),
    BorrowMutError(BorrowMutError),

    PidNotFound(Pid),
    ParseIntError(ParseIntError),
    RegexError(RegexError),
    IOError(IOError),
    FsError(more_fs::Error)
    // FsExtraError(FsError)
}

impl error::Error for Error {
    fn source(&self) -> Option<&(dyn error::Error + 'static)> {
        match self {
            Error::ParseIntError(e) => Some(e),
            Error::RegexError(e) => Some(e),
            Error::IOError(e) => Some(e),
            _ => None,
        }
    }
}

impl From<StdInfallible> for Error {
    fn from(o: StdInfallible) -> Error {
        Error::Infallible(o)
    }
}

impl From<BorrowError> for Error {
    fn from(o: BorrowError) -> Error {
        Error::BorrowError(o)
    }
}

impl From<BorrowMutError> for Error {
    fn from(o: BorrowMutError) -> Error {
        Error::BorrowMutError(o)
    }
}

impl From<NulError> for Error {
    fn from(o: NulError) -> Error {
        Error::CStringError(o)
    }
}

impl From<SerdeJsonError> for Error {
    fn from(o: SerdeJsonError) -> Error {
        Error::SerdeJsonError(o)
    }
}

impl From<ClapError> for Error {
    fn from(o: ClapError) -> Error {
        Error::ClapError(o)
    }
}

impl From<regex::Error> for Error {
    fn from(o: regex::Error) -> Error {
        Error::RegexError(o)
    }
}

// impl From<FsError> for Error {
//     fn from(o: FsError) -> Error {
//         Error::FsExtraError(o)
//     }
// }

impl From<more_fs::Error> for Error {
    fn from(o: more_fs::Error) -> Error {
        Error::FsError(o)
    }
}

impl From<ParseIntError> for Error {
    fn from(o: ParseIntError) -> Error {
        Error::ParseIntError(o)
    }
}

impl From<IOError> for Error {
    fn from(o: IOError) -> Error {
        Error::IOError(o)
    }
}

impl From<StripPrefixError> for Error {
    fn from(o: StripPrefixError) -> Error {
        Error::StripPrefixError(o)
    }
}

impl From<DkRegError> for Error {
    fn from(o: DkRegError) -> Error {
        Error::DockerRegistryError(o)
    }
}

impl fmt::Display for Error {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Error::CLIError(message) => {
                write!(f, "{}", message)
            }
            Error::NamespaceLinkNameError(message) => {
                write!(f, "NamespaceLinkNameError: {}", message)
            }
            Error::NamespaceParsingError(message) => {
                write!(f, "NamespaceParsingError: {}", message)
            }
            Error::NamespaceValidationError(message) => {
                write!(f, "NamespaceValidationError: {}", message)
            }
            Error::PidNotFound(pid) => write!(f, "PidNotFound: {}", pid.to_string()),
            Error::StateError(message) => write!(f, "state error: {}", message),
            _ => write!(f, "unhandled"),
        }
    }
}

#[derive(Debug, FromPrimitive, ToPrimitive)]
pub enum Errno {
    EPERM = 1,
    ENOENT,
    ESRCH,
    EINTR,
    EIO,
    ENXIO,
    E2BIG,
    ENOEXEC,
    EBADF,
    ECHILD,
    EAGAIN,
    ENOMEM,
    EACCES,
    EFAULT,
    ENOTBLK,
    EBUSY,
    EEXIST,
    EXDEV,
    ENODEV,
    ENOTDIR,
    EISDIR,
    EINVAL,
    ENFILE,
    EMFILE,
    ENOTTY,
    ETXTBSY,
    EFBIG,
    ENOSPC,
    ESPIPE,
    EROFS,
    EMLINK,
    EPIPE,
    EDOM,
    ERANGE,
}
