use crate::fs::fs_factory::FilesystemDesc;
use crate::types::Result;
use crate::errors::Error;

use crate::unix::mounts::{MountHandler, MountDesc, FsType, MountOpt};
use crate::unix::loop_handler::LoopHandler;

use std::io::Write;
use std::path::PathBuf;
use mktemp::Temp;

use std::process::Command;
use std::cmp;
use std::fs;

const BLOCK_SIZE: usize = 1024;
const BTRFS_MIN_SIZE: usize = BLOCK_SIZE * 1024 * 128;

/*
 * Descriptor for a btrfs filesystem
 */
pub struct BtrfsDesc {
    path: PathBuf,
    mount_desc: Option<MountDesc>
}

impl FilesystemDesc for BtrfsDesc {
    fn from_path(path: PathBuf) -> Result<Self> {
        Ok(BtrfsDesc{path, mount_desc: None})
    }
    fn from_mem(_data: Vec<u8>) -> Result<Self> {
        Err(Error::FilesystemDescError("no mem support for btrfs (forever?)".to_string()))
    }
    fn get_path(&self) -> PathBuf { self.path.clone() }
    
    fn format(&self, handler: &MountHandler, size: usize) -> Result<()> {
        // checking that the fs is not in use
        if handler.is_mounted(self.path.to_str().unwrap().into())? {
            return Err(Error::FilesystemDescError(
                format!("path {:?} already mounted", self.path.clone())))
        }

        // creating new image
        log::debug!("creating empty btrfs image on {:?}", self.path.clone());
        let temp_img_path = Temp::new_file()?.to_path_buf();        
        let mut temp_img = fs::OpenOptions::new()
            .read(false)
            .write(true)
            .append(false)
            .create(true)
            .open(temp_img_path.clone())?;

        let zero_vec = vec![0; BLOCK_SIZE];
        
        // inserting zero into the file
        for _ in 0..(cmp::max(size, BTRFS_MIN_SIZE) / BLOCK_SIZE + 1) {
            temp_img.write_all(zero_vec.as_slice())?;
        }

        // creating the filesystem
        log::debug!("formatting image as btrfs: {:?}", self.path.clone());
        let mut format_cmd = Command::new("mkfs.btrfs");
        format_cmd.arg(temp_img_path.clone());

        let mut format_process = format_cmd.spawn()?;
        let format_result = format_process.wait()?;

        if !format_result.success() {
            return Err(Error::FilesystemDescError(
                format!("could not format {:?} as btrfs: {}",
                        temp_img_path.clone().to_path_buf(),
                        format_result)));
        }

        // replacing old filesystem with the new one
        fs::copy(temp_img_path.clone(), self.path.clone())?;
        fs::remove_file(temp_img_path)?;
        
        Ok(())
    }

    fn mount(&mut self, handler: &MountHandler, loopctl: &LoopHandler, path: PathBuf)
             -> Result<()> {
        if !path.exists() {
            return Err(Error::FilesystemDescError("fs does not exist".into()))
        }

        // preparing the loop device
        let loopdev = loopctl.loop_file(self.path.clone())?;

        log::debug!("mounting btrfs {:?} on {:?}", loopdev.loop_dev, path.clone());
        // preparing the mount
        let temp_mount = MountDesc {
            source: loopdev.loop_dev.to_str().unwrap().into(),
            target: path.to_str().unwrap().into(),
            partition_type: FsType::Btrfs,
            options: vec![MountOpt::Noatime],
            data: None,
        };

        handler.mount(temp_mount.clone())?;

        log::debug!("mounting success: btrfs {:?} on {:?}", loopdev.loop_dev, path.clone());
        
        self.mount_desc = Some(temp_mount);

        Ok(())
    }

    fn umount(&mut self, handler: &MountHandler, loopctl: &LoopHandler) -> Result<()> {
        if self.mount_desc.is_none() {
            return Err(Error::FilesystemDescError("not mounted".into()));
        }

        // unmounting
        handler.umount(self.mount_desc.clone().unwrap(), false, false)?;

        // unlooping
        loopctl.unloop_file(self.path.clone())
    }
}
