use crate::types::Result;
use crate::errors::*;

use crate::unix::mounts::MountHandler;
use crate::unix::loop_handler::LoopHandler;

// use fs_extra;
use walkdir::WalkDir;
use num_traits::ToPrimitive;

use std::fs;
use std::path::PathBuf;
use std::os::unix::fs::*;

use mktemp::Temp;

use std::marker::PhantomData;

// // filesystem mount operations
// #[derive(Debug)]
// struct FsMount {
//     source: PathBuf,
//     target: PathBuf,
//     mount_type: String,
//     fs_type: String,
// }

// trait FsMountOp {
//     fn new(mounts: Vec<FsMount>) -> Self;
//     fn mount() -> Result<()>;
// }

// filesystem descriptor
pub trait FilesystemDesc {
    fn from_path(path: PathBuf) -> Result<Self> where Self: Sized;
    fn from_mem(data: Vec<u8>) -> Result<Self> where Self: Sized;
    fn format(&self, handler: &MountHandler, size: usize) -> Result<()>;

    fn get_path(&self) -> PathBuf;
    fn mount(&mut self, handler: &MountHandler, loopctl: &LoopHandler, path: PathBuf) -> Result<()>;
    fn umount(&mut self, handler: &MountHandler, loopctl: &LoopHandler) -> Result<()>;
}

// filesystem factory
pub struct FsFactory<I: FilesystemDesc, O: FilesystemDesc> {
    _i: PhantomData<I>,
    _o: PhantomData<O>,
    source_decided: bool,
    data: Option<Vec<u8>>,
    source: Option<PathBuf>,
    target: PathBuf
}

impl<I: FilesystemDesc, O: FilesystemDesc> FsFactory<I, O> {
    pub fn new(target: PathBuf) -> Self {
        FsFactory {
            _i: PhantomData,
            _o: PhantomData,
            source_decided: false,
            data: None,
            source: None,
            target
        }
    }
    
    pub fn set_data(&mut self, data: Vec<u8>) -> Result<()> {
        if self.source_decided {
            return Err(Error::FilesystemFactoryError(
                format!("source already decided for {:?}", self.target)))
        }

        self.source_decided = true;

        self.data = Some(data);

        Ok(())
    }
    
    pub fn set_source(&mut self, path: PathBuf) -> Result<()> {
        if self.source_decided {
            return Err(Error::FilesystemFactoryError(
                format!("source already decided for {:?}", self.target)))
        }

        if !path.is_file() {
            return Err(Error::FilesystemFactoryError(
                format!("source path {:?} does not exist", path)))
        }
        
        self.source_decided = true;

        self.source = Some(path);

        Ok(())        
    }

    // TODO: this should happen in a namespace, as copying files owned by root,
    // while keeping the permissions, is normally forbidden
    pub fn finish(&self) -> Result<O> {
        if !self.source_decided {
            return Err(Error::FilesystemFactoryError(
                format!("source for {:?} not decided", self.target)))
        }

        log::debug!("building fs descriptors");
        let mounts = &mut MountHandler::new(PathBuf::from("/")); // this root has to be configured
        let loopctl = &mut LoopHandler::new(None)?;
        
        let mut idesc = if self.data.is_some() {
            I::from_mem(self.data.as_ref().unwrap().to_owned())?
        } else {
            I::from_path(self.source.clone().unwrap())?
        };

        let temp_dir_input = Temp::new_dir()?.release();
        let temp_dir_output = Temp::new_dir()?.release();

        let odesc_size = if self.data.is_some() {
            self.data.as_ref().unwrap().len()
        } else {
            let input_metadata = fs::metadata(self.source.clone().unwrap())?;
            input_metadata.size().to_usize().unwrap()
        };

        let mut odesc = O::from_path(self.target.clone())?;
        odesc.format(mounts, odesc_size)?;

        log::debug!("mounting fs descriptors");
        idesc.mount(mounts, loopctl, temp_dir_input.clone())?;
        odesc.mount(mounts, loopctl, temp_dir_output.clone())?;

        // let mut copy_opts = fs_extra::dir::CopyOptions::new();
        // copy_opts.overwrite = true;
        // copy_opts.copy_inside = true;

        // let input_items = fs::read_dir(temp_dir_input.clone())?
        //     .into_iter()
        //     .map(|r| r.map(|di| di.path().to_str().unwrap().to_string()))
        //     .map(|r| r.map_err(Error::IOError))
        //     .collect::<Result<Vec<String>>>()?;

        log::debug!("copying data from {:?} to {:?}",
                    temp_dir_input.clone(), temp_dir_output.clone());

        // for input_item in input_items {
        //     let copied_path = &PathBuf::from(input_item);
        //     let elem_name = copied_path.file_name().unwrap();
        //     more_fs::copy_dir_all(copied_path, temp_dir_output.clone().join(elem_name))?;
        // }
       
        // fs_extra::copy_items(
        //     &input_items,
        //     &temp_dir_output.clone(),
        //     &copy_opts)?;

        copy_recursively(temp_dir_input.clone(), temp_dir_output.clone())?;
        
            // let temp_mount = MountDesc {
            //     source: "tmpfs".to_string(),
            //     target: temp_dir.to_str().unwrap().to_string(),
            //     partition_type: FsType::Tmpfs,
            //     options: vec![MountOpt::Noatime, MountOpt::Noexec],
            //     data: None,
        // };
        
        log::debug!("fs creation done, unmounting & exiting");

        idesc.umount(mounts, loopctl)?;
        odesc.umount(mounts, loopctl)?;
        
        Ok(odesc)
    }
}

fn copy_recursively(source: PathBuf, destination: PathBuf) -> Result<()> {
    for entry_r in WalkDir::new(source) {
        let entry = entry_r?;
        let metadata = entry.metadata()?;

        let rel_path = entry.path().strip_prefix(source)?;
        
        if entry.path_is_symlink() {
            
        } else if metadata.is_dir() {
            fs::DirBuilder::new()
                .mode(metadata.mode())
                .create(destination.clone().join(rel_path))?;
            fs::create_dir(destination.clone().join(rel_path))?;
        } else if metadata.is_file() {
            fs::copy(entry.path(), destination.clone().join(rel_path))?;
        } else {
            return Err(Error::FilesystemFactoryError(
                format!("unknown file type: {:?}", source)));
        }
    }
    Ok(())
}
