use crate::fs::fs_factory::FilesystemDesc;
use crate::types::Result;
use crate::errors::Error;

use crate::unix::mounts::{MountHandler, MountDesc, FsType, MountOpt};
use crate::unix::loop_handler::LoopHandler;

use std::path::PathBuf;
use mktemp::Temp;

use std::process::Command;
use std::fs;

/*
 * Descriptor for a (virtual & ephemeral) .tar.gz filesystem
 */
pub struct TargzFsDesc {
    path: PathBuf,
    is_from_mem: bool,
    mount_desc: Option<MountDesc>
}

impl FilesystemDesc for TargzFsDesc {
    fn from_path(path: PathBuf) -> Result<Self> {
        if !path.exists() {
            return Err(Error::FilesystemDescError("archive does not exist".into()))
        }
        Ok(TargzFsDesc{path, is_from_mem: false, mount_desc: None})
    }
    fn from_mem(data: Vec<u8>) -> Result<Self> {
        let mut temp_ar = Temp::new_path().release();
        temp_ar.set_extension("tar.gz");
        fs::write(temp_ar.clone(), data)?;
        
        Ok(TargzFsDesc{
            path: temp_ar,
            is_from_mem: true,
            mount_desc: None
        })
    }
    fn get_path(&self) -> PathBuf { self.path.clone() }
    
    fn format(&self, _handler: &MountHandler, _size: usize) -> Result<()> {
        Err(Error::FilesystemDescError("tar.gz archives cannot be formatted".into()))
    }

    fn mount(&mut self, handler: &MountHandler, _loopctl: &LoopHandler, path: PathBuf)
             -> Result<()> {
        let source_path_str = self.path
            .to_str()
            .map(Ok)
            .unwrap_or_else(|| Err(Error::FilesystemDescError("unable to parse path".into())))?;
        let target_path_str = path
            .to_str()
            .map(Ok)
            .unwrap_or_else(|| Err(Error::FilesystemDescError("unable to parse path".into())))?;

        // preparing the unpacking area
        let temp_mount = MountDesc {
            source: "tmpfs".to_string(),
            target: target_path_str.into(),
            partition_type: FsType::Tmpfs,
            options: vec![MountOpt::Noatime],
            data: None,
        };

        handler.mount(temp_mount.clone())?;

        // unpacking
        let mut tar_cmd = Command::new("tar");
        tar_cmd.arg("-xf");
        tar_cmd.arg(source_path_str);
        tar_cmd.arg("-C");
        tar_cmd.arg(target_path_str);

        let mut tar_process = tar_cmd.spawn()?;
        let unpack_result = tar_process.wait()?;

        if !unpack_result.success() {
            handler.umount(temp_mount.clone(), true, false)?;
            return Err(Error::FilesystemDescError("unpacking failed".into()));
        }

        self.mount_desc = Some(temp_mount);
        
        Ok(())
    }

    fn umount(&mut self, handler: &MountHandler, _loopctl: &LoopHandler)
              -> Result<()> {
        if self.mount_desc.is_none() {
            return Err(Error::FilesystemDescError("not mounted".into()));
        }

        // packing back the contents
        if !self.is_from_mem {
            // unpacking
            let mut temp_ar = Temp::new_path();
            temp_ar.set_extension("tar.gz");

            let mut tar_cmd = Command::new("tar");
            tar_cmd.arg("-a");
            tar_cmd.arg("-cf");
            tar_cmd.arg(self.path.clone());
            tar_cmd.arg("-C");
            tar_cmd.arg(self.mount_desc.clone().unwrap().target);
            tar_cmd.arg(".");

            let mut tar_process = tar_cmd.spawn()?;
            let pack_result = tar_process.wait()?;

            if !pack_result.success() {
                handler.umount(self.mount_desc.clone().unwrap(), true, false)?;
                return Err(Error::FilesystemDescError("unpacking failed".into()));
            }

            fs::copy(temp_ar.clone(), self.path.clone())?;
            fs::remove_file(temp_ar)?;
        }

        // unmounting the tar
        handler.umount(self.mount_desc.clone().unwrap(), false, false)?;
        self.mount_desc = None;
        
        Ok(())
    }
}
