use super::namespace_type::NamespaceType;

use crate::errors::Error;
use crate::types::{NsId, Result};

use serde::Serialize;

use regex::{Captures, Regex};

use std::str::FromStr;

#[derive(Debug, Clone, Copy, Hash, PartialEq, Eq, Serialize)]
pub enum Namespace {
    Cgroup(NsId),
    Ipc(NsId),
    Mnt(NsId),
    Net(NsId),
    Pid(NsId),
    Time(NsId),
    User(NsId),
    Uts(NsId),
}

impl Namespace {
    pub fn get_id(&self) -> &NsId {
        match self {
            Namespace::Cgroup(id) => id,
            Namespace::Ipc(id) => id,
            Namespace::Mnt(id) => id,
            Namespace::Net(id) => id,
            Namespace::Pid(id) => id,
            Namespace::Time(id) => id,
            Namespace::User(id) => id,
            Namespace::Uts(id) => id,
        }
    }

    pub fn get_type(&self) -> NamespaceType {
        match self {
            Namespace::Cgroup(_) => NamespaceType::Cgroup,
            Namespace::Ipc(_) => NamespaceType::Ipc,
            Namespace::Mnt(_) => NamespaceType::Mnt,
            Namespace::Net(_) => NamespaceType::Net,
            Namespace::Pid(_) => NamespaceType::Pid,
            Namespace::Time(_) => NamespaceType::Time,
            Namespace::User(_) => NamespaceType::User,
            Namespace::Uts(_) => NamespaceType::Uts,
        }
    }
}

impl ToString for Namespace {
    fn to_string(&self) -> String {
        format!("{}[{}]", self.get_type().to_string(), self.get_id())
    }
}

impl FromStr for Namespace {
    type Err = Error;

    fn from_str(s: &str) -> Result<Namespace> {
        // lazy_static! {
        //     static ref re: Regex = new(r"^([a-z]+):\[([0-9]+)\]$")?;
        // }
        let re: Regex = Regex::new(r"^([a-z]+):\[([0-9]+)\]$")?;

        let captures: Captures = re
            .captures(s)
            .ok_or(Error::NamespaceParsingError(format!(
                "unable to run regex for {}",
                s
            )))
            .and_then(|c: Captures| match c.len() {
                3 => Ok(c),
                _ => Err(Error::NamespaceParsingError(format!(
                    "unable to find all fields for {}",
                    s
                ))),
            })?;

        let ns_type_lit = captures.get(1).unwrap().as_str();
        let ns_id_lit = captures.get(2).unwrap().as_str();

        let ns_type = NamespaceType::from_str(ns_type_lit)?;
        let ns_id = ns_id_lit.parse::<NsId>()?;

        let ns = ns_type.apply(ns_id);
        Ok(ns)
    }
}
