use super::namespace::Namespace;

use serde::Serialize;

use crate::errors::Error;
use crate::types::{NsId, Result};

use std::str::FromStr;
use std::string::ToString;

#[derive(Debug, Copy, Clone, PartialEq)]
pub enum NamespaceType {
    Cgroup = 0x02000000,
    Ipc = 0x08000000,
    Mnt = 0x00020000,
    Net = 0x40000000,
    Pid = 0x20000000,
    Time = 0x00000080,
    User = 0x10000000,
    Uts = 0x04000000,
}

impl NamespaceType {
    pub fn apply(&self, ns_id: NsId) -> Namespace {
        match self {
            NamespaceType::Cgroup => Namespace::Cgroup(ns_id),
            NamespaceType::Ipc => Namespace::Ipc(ns_id),
            NamespaceType::Mnt => Namespace::Mnt(ns_id),
            NamespaceType::Net => Namespace::Net(ns_id),
            NamespaceType::Pid => Namespace::Pid(ns_id),
            NamespaceType::Time => Namespace::Time(ns_id),
            NamespaceType::User => Namespace::User(ns_id),
            NamespaceType::Uts => Namespace::Uts(ns_id),
        }
    }
}

impl ToString for NamespaceType {
    fn to_string(&self) -> String {
        match self {
            NamespaceType::Cgroup => "cgroup",
            NamespaceType::Ipc => "ipc",
            NamespaceType::Mnt => "mnt",
            NamespaceType::Net => "net",
            NamespaceType::Pid => "pid",
            NamespaceType::Time => "time",
            NamespaceType::User => "user",
            NamespaceType::Uts => "uts",
        }
        .to_string()
    }
}

impl FromStr for NamespaceType {
    type Err = Error;

    fn from_str(s: &str) -> Result<NamespaceType> {
        match s {
            "cgroup" => Ok(NamespaceType::Cgroup),
            "ipc" => Ok(NamespaceType::Ipc),
            "mnt" => Ok(NamespaceType::Mnt),
            "net" => Ok(NamespaceType::Net),
            "pid" => Ok(NamespaceType::Pid),
            "time" => Ok(NamespaceType::Time),
            "user" => Ok(NamespaceType::User),
            "uts" => Ok(NamespaceType::Uts),
            _ => Err(Error::NamespaceParsingError(
                "undefined namespace type".to_string(),
            )),
        }
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash, Serialize)]
pub enum NamespaceOption {
    Cgroup,
    Ipc,
    Mnt,
    Net,
    Pid,
    PidForChildren,
    Time,
    TimeForChildren,
    User,
    Uts,
}

impl ToString for NamespaceOption {
    fn to_string(&self) -> String {
        match self {
            NamespaceOption::Cgroup => "cgroup",
            NamespaceOption::Ipc => "ipc",
            NamespaceOption::Mnt => "mnt",
            NamespaceOption::Net => "net",
            NamespaceOption::Pid => "pid",
            NamespaceOption::PidForChildren => "pid_for_children",
            NamespaceOption::Time => "time",
            NamespaceOption::TimeForChildren => "time_for_children",
            NamespaceOption::User => "user",
            NamespaceOption::Uts => "uts",
        }
        .to_string()
    }
}

impl FromStr for NamespaceOption {
    type Err = Error;

    fn from_str(s: &str) -> Result<NamespaceOption> {
        match s {
            "cgroup" => Ok(NamespaceOption::Cgroup),
            "ipc" => Ok(NamespaceOption::Ipc),
            "mnt" => Ok(NamespaceOption::Mnt),
            "net" => Ok(NamespaceOption::Net),
            "pid" => Ok(NamespaceOption::Pid),
            "pid_for_children" => Ok(NamespaceOption::PidForChildren),
            "time" => Ok(NamespaceOption::Time),
            "time_for_children" => Ok(NamespaceOption::TimeForChildren),
            "user" => Ok(NamespaceOption::User),
            "uts" => Ok(NamespaceOption::Uts),
            _ => Err(Error::NamespaceParsingError(
                "undefined namespace option".to_string(),
            )),
        }
    }
}
