use serde::de::DeserializeOwned;
use serde::Serialize;

use crate::errors::Error;
use crate::types::Result;

use std::fs;
use std::path::PathBuf;

use std::thread;
use std::time::Duration;

const LOCK_FILENAME: &str = ".lock";
const RLOCK_FILENAME: &str = ".rlock";
const TYPE_FILENAME: &str = ".state_type";

const LOOP_WAIT_MS: u64 = 100;

/*
 * Dummy state to handle uninitialized state dirs
 */
struct DummyState {
    base_path: PathBuf
}

impl AbstractState for DummyState {
    fn open_internal(_: PathBuf, _: Option<DummyState>) -> Result<DummyState> {
        Err(Error::StateError("DummyState cannot be instantiated".to_string()))
    }
    
    fn desc() -> String {
        "".to_string()
    }

    fn empty_variant() -> DummyState {
        DummyState {
            base_path: PathBuf::new()
        }
    }

    fn base_path(&self) -> PathBuf {
        self.base_path.clone()
    }
    
    fn state_path(&self) -> PathBuf {
        PathBuf::new()
    }
}

/*
 * Abstract state
 */
pub trait AbstractState {
    // static methods
    fn open_internal(path: PathBuf, init: Option<Self>) -> Result<Self>
    where Self: Sized;
    
    fn desc() -> String;

    fn empty_variant() -> Self
    where Self: Sized;
    
    // required methods
    fn base_path(&self) -> PathBuf;
    fn state_path(&self) -> PathBuf;

    /*
     * initialization and opening
     */
    fn open(path: PathBuf, init: Option<Self>) -> Result<Self>
    where Self: Sized {
        log::debug!("opening (init? {}) at {}", init.is_some(), path.to_string_lossy());

        // checking if path is absolute
        if path.is_relative() {
            return Err(Error::StateError(
                format!("the path {} is relative", path.to_string_lossy())))
        }
        
        // checking if directory exists
        if init.is_some() {
            fs::create_dir_all(&path)?;
        } else {
            fs::read_dir(&path)?;
        }

        let dummy = DummyState{base_path: path.clone()};
        
        dummy.wait_for_lock()?;
        dummy.lock()?;

        let type_file_path = path.join(TYPE_FILENAME);

        // checking state type
        if init.is_some() && !type_file_path.exists() {
            log::debug!("creating state type file for {}", Self::desc());

            fs::write(type_file_path, Self::desc())
                .or_else(|e| {
                    log::debug!("error reading state type file: {}", e);
                    dummy.unlock()?;
                    Err(Error::IOError(e))
                })?;
        } else {
            if init.is_some() {
                log::warn!("initializing {} on {}, state file already exists",
                           Self::desc(), type_file_path.to_string_lossy())
            }
            
            log::debug!("opening state type file for {}", Self::desc());

            let found_type = fs::read_to_string(type_file_path)
                .or_else(|e| {
                    log::debug!("error reading state type file: {}", e);
                    dummy.unlock()?;
                    Err(Error::IOError(e))
                })?;

            if found_type.to_owned() != Self::desc() {
                dummy.unlock()?;
                return Err(Error::StateError(
                    format!("type {} does not match required type {}", found_type, Self::desc())));
            }
        }

        let s = Self::open_internal(path, init);
        dummy.unlock()?;
        s
    }
    
    fn get_item<T: AbstractState>(&self, item_state_path: PathBuf) -> Result<T> {
        let rel_item_path = match item_state_path.is_relative() {
            true => Ok(item_state_path),
            false => item_state_path
                .strip_prefix(self.state_path())
                .map(|p| p.to_path_buf()),
        }?;

        let abs_item_path = self.base_path().join(rel_item_path);

        if !abs_item_path.is_dir() {
            return Err(Error::StateGetItemError("could not find item".to_string()));
        }

        T::open(abs_item_path, None)
    }

    /*
     * attributes
     */
    fn write_attr<M: Serialize>(&self, attr: String, data: M) -> Result<()> {
        self.lock()?;
        fs::write(self.base_path().join(attr), serde_json::to_vec_pretty(&data)?)?;
        self.unlock()?;
        Ok(())
    }

    fn read_attr<M: DeserializeOwned>(&self, attr: String) -> Result<M> {
        self.lock()?;
        let raw_data = fs::read(self.base_path().join(attr))?;
        let data = serde_json::from_slice(raw_data.as_slice())?;
        self.unlock()?;
        Ok(data)
    }
    
    /*
     * attributes (raw)
     */
    fn write_attr_raw(&self, attr: String, data: Vec<u8>) -> Result<()> {
        self.lock()?;
        fs::write(self.base_path().join(attr), data)?;
        self.unlock()?;
        Ok(())
    }

    fn read_attr_raw(&self, attr: String) -> Result<Vec<u8>> {
        self.lock()?;
        let data = fs::read(self.base_path().join(attr))?;
        self.unlock()?;
        Ok(data)
    }
    
    /*
     * locking functions
     */
    fn lock(&self) -> Result<()> {
        let base_path = self.base_path();
        if self.is_locked()? {
            return Err(Error::StateError(format!(
                "state {} already locked",
                base_path.to_str().unwrap()
            )));
        }

        fs::File::create(base_path.join(LOCK_FILENAME))?;
        Ok(())
    }

    fn unlock(&self) -> Result<()> {
        let base_path = self.base_path();
        if !self.is_locked()? {
            return Err(Error::StateError(format!(
                "state {} already unlocked",
                base_path.to_string_lossy()
            )));
        }

        fs::remove_file(base_path.join(LOCK_FILENAME))?;
        Ok(())
    }

    fn is_locked(&self) -> Result<bool> {
        Ok(self.base_path().join(LOCK_FILENAME).exists() ||
           self.base_path().join(RLOCK_FILENAME).exists())
    }

    fn wait_for_lock(&self) -> Result<()> {
        while self.is_locked()? {
            thread::sleep(Duration::from_millis(LOOP_WAIT_MS))
        }
        Ok(())
    }

    /*
     * read locking functions
     */
    fn rlock(&self) -> Result<()> {
        let lock_path = self.base_path().join(RLOCK_FILENAME);
        if self.is_rlocked()? {
            return Err(Error::StateError(format!(
                "state {} already locked",
                self.base_path().to_string_lossy()
            )));
        }

        let current_count = fs::read_to_string(lock_path.clone())
            .map_err(Error::IOError)
            .and_then(|lock_count_str: String| -> Result<u128> {
                lock_count_str
                    .parse::<u128>()
                    .map_err(Error::ParseIntError)
            })
            .or::<Error>(Ok(0))?;

        fs::write(lock_path, current_count.to_string().as_bytes())?;
        Ok(())
    }

    fn unrlock(&self) -> Result<()> {
        let lock_path = self.base_path().join(RLOCK_FILENAME);

        if !lock_path.exists() {
            return Err(Error::StateError(format!(
                "state {} already unlocked",
                self.base_path().to_string_lossy()
            )));
        }
        
        let current_count = fs::read_to_string(lock_path.clone())
            .map_err(Error::IOError)
            .and_then(|lock_count_str: String| -> Result<u128> {
                lock_count_str
                    .parse::<u128>()
                    .map_err(Error::ParseIntError)
            })
            .or::<Error>(Ok(0))?;

        if current_count < 2 {
            fs::remove_file(lock_path.clone())?;
        } else {
            fs::write(lock_path, (current_count - 1).to_string())?;
        }
        Ok(())
    }

    fn is_rlocked(&self) -> Result<bool> {
        Ok(self.base_path().join(LOCK_FILENAME).exists())
    }

    fn wait_for_rlock(&self) -> Result<()> {
        while self.is_rlocked()? {
            thread::sleep(Duration::from_millis(LOOP_WAIT_MS))
        }
        Ok(())
    }

}
