use crate::state::AbstractState;
use crate::types::Result;

use std::borrow::BorrowMut;
use std::path::PathBuf;
use std::fs;

use serde::*;

use sha2::{Sha512, Digest};
use base58::*;

use std::str::FromStr;

const FS_STORE_TYPE: &str = "fs_store";
const FS_STORE_ITEM_PREFIX: &str = "item_";
const FS_STORE_METADATA_FILENAME: &str = "metadata";
const FS_STORE_IMAGE_FILENAME: &str = "image";

/*
 * Filesystem item
 */
#[derive(Debug, Clone, Serialize, Deserialize)]
pub enum FsItemType {
    Ext4,
    Fat32,
    Btrfs
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub enum FsItemEncryption {}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct FsStoreItemId {
    pub id: String,
}

#[derive(Hash, Debug, Clone, Serialize, Deserialize)]
pub struct FsStoreItemName {
    pub domain: String,
    pub name: String,
    pub version: String,
    pub overlay: u64,
}

impl FsStoreItemName {
    fn hash(&self) -> String {
        let mut hasher = Sha512::default();
        hasher.update(self.domain.clone());
        hasher.update(self.name.clone());
        hasher.update(self.version.clone());
        hasher.update(self.overlay.to_string());
        let hashed_data = hasher.finalize();
        hashed_data.as_slice().to_base58()
    }
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct FsStoreItem {
    pub item_name: FsStoreItemName,
    pub item_hash: FsStoreItemId,
    pub prev_overlay: Option<String>,
    pub fs_type: FsItemType,
    pub encryption: Option<FsItemEncryption>,
    pub contents_checksum: Option<String>,
    #[serde(skip_serializing, skip_deserializing)]
    pub base_path: PathBuf
}

impl AbstractState for FsStoreItem {
    fn desc() -> String { "fs_store_item".to_string() }

    fn empty_variant() -> FsStoreItem {
        FsStoreItem {
            item_name: FsStoreItemName {
                domain: "".to_string(),
                name: "".to_string(),
                version: "".to_string(),
                overlay: 0
            },
            item_hash: FsStoreItemId {
                id: "".to_string()
            },
            prev_overlay: None,
            fs_type: FsItemType::Btrfs,
            encryption: None,
            contents_checksum: None,
            base_path: PathBuf::new()
        }
    }
    
    fn base_path(&self) -> PathBuf {
        self.base_path.clone()
    }

    fn state_path(&self) -> PathBuf {
        PathBuf::from_str("/").unwrap()
    }
    
    fn open_internal(path: PathBuf, init: Option<FsStoreItem>) -> Result<FsStoreItem>
    where Self: Sized {
        let metadata_filename = path.join(FS_STORE_METADATA_FILENAME);

        if init.is_none() {
            let metadata_raw = fs::read_to_string(metadata_filename)?;
            let read_item: FsStoreItem = serde_json::from_str(metadata_raw.as_str())?;
            return Ok(read_item);
        } else {
            let mut new_item = init.unwrap();
            new_item.reset_id();
            new_item.base_path = path;
            // this would result in a deadlock i guess... for now
            // new_item.commit_metadata()?;
            Ok(new_item)
        }
    }
}

impl FsStoreItem {
    pub fn reset_id(&mut self) {
        self.item_hash.id = self.item_name.hash();
    }

    pub fn commit_metadata(&self) -> Result<()> {
        self.write_attr(FS_STORE_METADATA_FILENAME.to_string(), self.clone())
    }

    pub fn write_image(&self, data: Vec<u8>) -> Result<()> {
        self.write_attr_raw(FS_STORE_IMAGE_FILENAME.into(), data)
    }

    pub fn read_image(&self) -> Result<Vec<u8>> {
        self.read_attr_raw(FS_STORE_IMAGE_FILENAME.into())
    }
}

/*
 * Filesystem store
 */

pub struct FsStore {
    base_path: PathBuf,
    fs_items: Vec<FsStoreItem>,
}

impl AbstractState for FsStore {
    fn desc() -> String { FS_STORE_TYPE.to_string() }
    fn base_path(&self) -> PathBuf { self.base_path.clone() }
    fn state_path(&self) -> PathBuf { self.base_path.clone() }

    fn empty_variant() -> FsStore {
        FsStore {
            base_path: PathBuf::new(),
            fs_items: Vec::new()
        }
    }
    
    fn open_internal(path: PathBuf, _init: Option<FsStore>) -> Result<FsStore> {
        let mut found_fs_items = Vec::new();
        
        for found_item in fs::read_dir(path.clone())? {
            let found_item = found_item?;

            let found_item_path = found_item.path();
            let found_item_name = found_item.file_name().into_string();

            if found_item_name.is_err() {
                log::warn!("malformed item: {}", found_item.file_name().to_string_lossy());
                continue;
            }

            let found_item_name = found_item_name.unwrap();
            
            if found_item_path.is_dir() && found_item_name.starts_with(FS_STORE_ITEM_PREFIX) {
                let new_item = FsStoreItem::open(path.clone().join(found_item_name), None)?;
                found_fs_items.push(new_item);
            }
        }

        Ok(FsStore {
            base_path: path,
            fs_items: found_fs_items
        })
    }
}

impl FsStore {
    pub fn add_item(&mut self, item: &mut FsStoreItem) -> Result<()> {
        item.reset_id();
        
        let new_item_dirname = format!("{}{}", FS_STORE_ITEM_PREFIX, item.item_hash.id.clone());
        let new_item_path = self.base_path().join(new_item_dirname);

        log::debug!("adding item {:?} to {:?}", item, new_item_path);
        
        self.lock()?;
        let mut new_item = FsStoreItem::open(new_item_path, Some(item.clone()))?;
        new_item.commit_metadata()?;
        self.unlock()?;
        
        self.fs_items.push(new_item.clone());

        item.clone_from(new_item.borrow_mut());
        
        Ok(())
    }
}
