use crate::errors::Error;
use crate::namespace::{Namespace, NamespaceOption};
use crate::types::{Expectable, Pid, Result};

use std::fs;
use std::io::prelude::*;
use std::path::{Path, PathBuf};

// use std::boxed::Box;
// use std::rc::Rc;

use std::str::FromStr;

use log;

static COMM: &str = "comm";
static CMDLINE: &str = "cmdline";

static PROCDIR: &str = "/proc";

pub struct ProcHandler {
    root_dir: PathBuf,
}

impl ProcHandler {
    pub fn new(root_dir: PathBuf) -> Self {
        ProcHandler { root_dir }
    }

    fn proc_path(&self) -> PathBuf {
        Path::new(&self.root_dir).join(Path::new(PROCDIR))
    }

    pub fn pid_path(&self, pid: Pid) -> PathBuf {
        self.proc_path().join(pid.pid.to_string())
    }

    pub fn ns_path(&self, pid: u64, ns_opt: NamespaceOption) -> PathBuf {
        self.proc_path()
            .join(Path::new(&pid.clone().to_string()))
            .join(Path::new("ns"))
            .join(Path::new(&ns_opt.to_string()))
    }

    fn check_pid(&self, pid: Pid) -> Result<()> {
        let pid_path = self.proc_path().join(Path::new(&pid.to_string()));

        fs::read_dir(&pid_path).or_else(|_| Err(Error::PidNotFound(pid)))?;
        Ok(())
    }

    pub fn get_pids(&self) -> Result<Vec<Pid>> {
        let proc_dirs: fs::ReadDir = fs::read_dir(self.proc_path())?;

        let pids: Vec<Pid> = proc_dirs
            .flat_map(|entry| entry.map(|entry: fs::DirEntry| entry.file_name().into_string()))
            .flatten()
            .filter_map(|fname: String| {
                fname
                    .parse::<u64>()
                    .map(|parsed_pid| Some(parsed_pid))
                    .unwrap_or(None)
            })
            .flat_map(|pid: u64| -> Result<Pid> {
                let comm_path = self.proc_path().join(pid.to_string()).join(COMM);
                let cmdline_path = self.proc_path().join(pid.to_string()).join(CMDLINE);

                let mut desc = String::new();

                let mut cmdline_file = fs::File::open(cmdline_path.clone())
                    .spect(format!("unable to open cmdfile for pid {}", pid))?;
                cmdline_file.read_to_string(&mut desc)?;

                if desc.is_empty() {
                    let mut comm_file = fs::File::open(comm_path.clone())
                        .spect(format!("unable to open comm file for pid {}", pid))?;
                    comm_file.read_to_string(&mut desc)?;
                }

                Ok(Pid {
                    pid,
                    desc: desc.replace("\u{0}", " ").trim().to_owned(),
                })
            })
            .collect();

        for p in &pids {
            log::debug!("found PID {:?}", p);
        }

        Ok(pids)
    }

    pub fn get_namespace(&self, pid: Pid, ns_opt: NamespaceOption) -> Result<Namespace> {
        self.check_pid(pid.clone())?;

        let ns_path = self.ns_path(pid.pid, ns_opt);

        let ns_link = fs::read_link(&ns_path).spect(format!(
            "namespace {} does not exist for pid {}",
            ns_opt.to_string(),
            pid.pid
        ))?;

        let ns_fqn = ns_link
            .file_name()
            .and_then(|name| name.to_str())
            .map(|name_str: &str| name_str.to_owned())
            .ok_or(Error::NamespaceLinkNameError(
                ns_path
                    .to_str()
                    .map(|raw_name| raw_name.to_owned())
                    .unwrap_or(".".to_owned()),
            ))?;

        Namespace::from_str(&ns_fqn)
    }
}
