use crate::types::Result;
use crate::errors::{Error, Errno};

use syscalls::*;

use num_traits::FromPrimitive;

use std::fs;
use std::ops::Index;
use std::os::unix::prelude::IntoRawFd;
use std::path::PathBuf;
use std::cell::RefCell;

const LOOP_CTL_DEFAULT_PATH: &str = "/dev/loop-control";
const LOOP_CTL_GET_FREE: i64 = 0x4C82;
const LOOP_SET_FD: i64 = 0x4C00;

#[derive(Debug, Clone)]
pub struct LoopAssoc {
    pub loop_no: i64,
    pub loop_file: PathBuf,
    pub loop_dev: PathBuf
}

#[derive(Debug, Clone)]
pub struct LoopHandler {
    control: PathBuf,
    associations: RefCell<Vec<LoopAssoc>>
}


impl LoopHandler {
    pub fn new(control_path_arg: Option<PathBuf>) -> Result<Self> {
        let control_path = control_path_arg.unwrap_or(PathBuf::from(LOOP_CTL_DEFAULT_PATH));
        
        let control_loop = fs::OpenOptions::new()
            .read(true)
            .write(true)
            .create(false)
            .open(control_path.clone())
            .map_err(|e| {
                Error::LoopCtrlError(
                    format!("unable to open loop control: {}", e))
            })?;

        let control_loop_fd = control_loop.into_raw_fd();
        let free_result = unsafe { syscall!(SYS_ioctl, control_loop_fd, LOOP_CTL_GET_FREE) };

        if free_result.is_err() {
            return Err(Error::LoopCtrlError(
                format!("unable to allocate free loop: {:?}",
                        Errno::from_i64(free_result.err().unwrap()))));
        }
        
        Ok(LoopHandler{
            control: control_path.clone(),
            associations: RefCell::new(Vec::new())
        })
    }

    pub fn loop_file(&self, filepath: PathBuf) -> Result<LoopAssoc> {
        log::debug!("mounting {:?} as loop", filepath.clone());
        let control_loop = fs::OpenOptions::new()
            .read(true)
            .write(true)
            .create(false)
            .open(self.control.clone())
            .map_err(|e| {
                Error::LoopCtrlError(
                    format!("unable to open loop control: {}", e))
            })?;

        log::debug!("allocating new loop device");
        let control_loop_fd = control_loop.into_raw_fd();
        let free_result = unsafe { syscall!(SYS_ioctl, control_loop_fd, LOOP_CTL_GET_FREE) };

        if free_result.is_err() {
            return Err(Error::LoopCtrlError(
                format!("unable to allocate free loop: {:?}",
                        Errno::from_i64(free_result.err().unwrap()))));
        }

        let free_loop_dev_no = free_result.ok().unwrap();
        let free_loop_dev_path = self.control.clone()
            .parent()
            .unwrap_or(PathBuf::from("/").as_path())
            .join(PathBuf::from(format!("loop{}", free_loop_dev_no)));

        log::debug!("opening loop device {:?}", free_loop_dev_path.clone());
        let free_loop_dev = fs::OpenOptions::new()
            .read(true)
            .write(true)
            .create(false)
            .open(free_loop_dev_path.clone())
            .map_err(|e| {
                Error::LoopCtrlError(
                    format!("unable to open loop: {}", e))
            })?;

        let image_file = fs::OpenOptions::new()
            .read(true)
            .write(true)
            .create(false)
            .open(filepath.clone())
            .map_err(|e| {
                Error::LoopCtrlError(
                    format!("unable to open file image: {}", e))
            })?;
        

        let free_loop_fd = free_loop_dev.into_raw_fd();
        let image_file_fd = image_file.into_raw_fd();

        log::debug!("looping {:?} on {:?}", filepath.clone(), free_loop_dev_path.clone());
        let loop_result = unsafe { syscall!(SYS_ioctl, free_loop_fd, LOOP_SET_FD, image_file_fd) };

        if loop_result.is_err() {
            return Err(Error::LoopCtrlError(
                format!("unable to loop image: {:?}",
                        Errno::from_i64(loop_result.err().unwrap()))));
        }

        log::debug!("looping success: {:?} on {:?}", filepath.clone(), free_loop_dev_path.clone());
        let new_loop_assoc = LoopAssoc {
            loop_dev: free_loop_dev_path.clone(),
            loop_file: filepath.clone(),
            loop_no: free_loop_dev_no
        };

        self.associations.borrow_mut().push(new_loop_assoc.clone());
        
        Ok(new_loop_assoc)
    }

    pub fn unloop_file(&self, filepath: PathBuf) -> Result<()> {
        let associations = self.associations.borrow();
        let associations_iter = associations.iter();
        let found_loop_assoc = associations_iter
            .filter(|assoc| assoc.loop_file == filepath.clone())
            .collect::<Vec<&LoopAssoc>>();
        let loop_assoc = found_loop_assoc.first();
        
        if loop_assoc.is_none() {
                return Err(Error::LoopDeviceError(
                    format!("loop path {:?} is not mounted or managed", filepath.clone())))
        }

        log::debug!("unlooping {:?} on {:?}", filepath.clone(),
                    loop_assoc.unwrap().loop_dev.clone());
        let disassoc_result =
            unsafe { syscall!(SYS_ioctl, loop_assoc.unwrap().loop_no) };

        if disassoc_result.is_err() {
            return Err(Error::LoopCtrlError(
                format!("unable to disassociate loop dev: {:?}",
                        Errno::from_i64(disassoc_result.err().unwrap()))));
        }
        log::debug!("unlooping success: {:?} on {:?}",
                    filepath.clone(), loop_assoc.unwrap().loop_dev.clone());

        // remove loop association
        // self.associations.borrow_mut().
        Ok(())
    }
}
