use crate::errors::{Errno, Error};
use crate::types::Result;

use num_traits::FromPrimitive;
use syscalls::*;

use std::ffi::CString;
use std::fs;
use std::path::{Path, PathBuf};
use std::str::FromStr;

static PROCDIR: &str = "/proc";
static MOUNTS: &str = "mounts";

#[derive(PartialEq, Clone, Debug)]
pub enum FsType {
    Ext2,
    Ext3,
    Ext4,
    Vfat,
    Btrfs,
    Tmpfs,
    Proc,
    Sysfs,
    Devtmpfs,
    Devpts,
    Mqueue,
    Tracefs,
    Cgroup,
    NoneFs,
    Other(String),
}

impl FromStr for FsType {
    type Err = Error;

    fn from_str(o: &str) -> Result<FsType> {
        Ok(match o {
            "ext2" => FsType::Ext2,
            "ext3" => FsType::Ext3,
            "ext4" => FsType::Ext4,
            "vfat" => FsType::Vfat,
            "btrfs" => FsType::Btrfs,
            "tmpfs" => FsType::Tmpfs,
            "proc" => FsType::Proc,
            "sysfs" => FsType::Sysfs,
            "devtmpfs" => FsType::Devtmpfs,
            "devpts" => FsType::Devpts,
            "mqueue" => FsType::Mqueue,
            "tracefs" => FsType::Tracefs,
            "cgroup" => FsType::Cgroup,
            part_type => FsType::Other(part_type.to_string()),
        })
    }
}

impl ToString for FsType {
    fn to_string(&self) -> String {
        match &self {
            FsType::Ext2 => "ext2",
            FsType::Ext3 => "ext3",
            FsType::Ext4 => "ext4",
            FsType::Vfat => "vfat",
            FsType::Tmpfs => "tmpfs",
            FsType::Btrfs => "btrfs",
            FsType::Proc => "proc",
            FsType::Sysfs => "sysfs",
            FsType::Devtmpfs => "devtmpfs",
            FsType::Devpts => "devpts",
            FsType::Mqueue => "mqueue",
            FsType::Tracefs => "tracefs",
            FsType::Cgroup => "cgroup",
            FsType::Other(part_type) => part_type,
            FsType::NoneFs => "",
        }
        .to_string()
    }
}

#[derive(Clone, Debug)]
pub enum MountOpt {
    Rdonly = 1,
    Nosuid = 2,
    Nodev = 4,
    Noexec = 8,
    Noatime = 1024,
    Relatime = 1 << 21,

    Bind = 4096,
    Unbindable = 1 << 17,
    Private = 1 << 18,
    Slave = 1 << 19,
    Shared = 1 << 20,

    Remount = 32,
}

struct MountOptRead(Option<MountOpt>);

impl FromStr for MountOptRead {
    type Err = Error;

    fn from_str(o: &str) -> Result<MountOptRead> {
        Ok(MountOptRead(match o {
            "noatime" => Some(MountOpt::Noatime),
            "relatime" => Some(MountOpt::Relatime),
            "noexec" => Some(MountOpt::Noexec),
            "nodev" => Some(MountOpt::Nodev),
            "nosuid" => Some(MountOpt::Nosuid),
            "ro" => Some(MountOpt::Rdonly),
            _ => None,
        }))
    }
}

#[derive(Clone, Debug)]
pub struct MountDesc {
    pub source: String,
    pub target: String,
    pub partition_type: FsType,
    pub options: Vec<MountOpt>,
    pub data: Option<String>,
}

#[derive(Clone)]
pub struct MountHandler {
    pub root_dir: PathBuf,
}

impl MountHandler {
    pub fn new(root_dir: PathBuf) -> Self {
        MountHandler { root_dir }
    }

    fn proc_path(&self) -> PathBuf {
        Path::new(&self.root_dir).join(Path::new(PROCDIR))
    }

    pub fn get_mounts(&self) -> Result<Vec<MountDesc>> {
        let mounts_file = self.proc_path().join(MOUNTS);
        let current_mounts = fs::read_to_string(mounts_file)?;

        let mut mount_descriptors = vec![];

        for mount_line in current_mounts.split_terminator("\n").into_iter() {
            let mount_line_items: Vec<String> =
                mount_line.split_whitespace().map(str::to_string).collect();

            let source = mount_line_items[0].clone();
            let target = mount_line_items[1].clone();
            let partition_type = FsType::from_str(mount_line_items[2].as_str())?;
            let options = mount_line_items[3]
                .split(",")
                .map(MountOptRead::from_str)
                .collect::<Result<Vec<MountOptRead>>>()?
                .iter()
                .map(|mor| mor.0.clone())
                .flatten()
                .collect();

            mount_descriptors.append(&mut vec![MountDesc {
                source,
                target,
                partition_type,
                options,
                data: None,
            }])
        }

        Ok(mount_descriptors)
    }

    pub fn is_mounted(&self, path: String) -> Result<bool> {
        let current_mounts = self.get_mounts()?;
        Ok(current_mounts
           .iter()
           .find(|mount_desc| {
               let source_path = mount_desc.source.clone();
               let target_path = mount_desc.target.clone();
               path == source_path || path == target_path
           })
           .is_some())
    }
    
    pub fn umount(&self, mount_desc: MountDesc, force: bool, lazy: bool) -> Result<()> {
        cpointer_string!(mnt_target_cptr, mnt_target_cstring, mount_desc.target);

        let umount_flags = (force as i64) + (lazy as i64) * 2;

        log::debug!("unmounting {}", mount_desc.target);

        let syscall_result = unsafe { syscall!(SYS_umount2, mnt_target_cptr, umount_flags) };

        syscall_result.map(|_| {}).map_err(|errno| {
            Error::ProviderError(format!("umount2, errno: {:?}", Errno::from_i64(errno)))
        })
    }

    pub fn mount(&self, mount_desc: MountDesc) -> Result<()> {
        cpointer_string!(mnt_source_cptr, mnt_source_cstring, mount_desc.source);
        cpointer_string!(mnt_target_cptr, mnt_target_cstring, mount_desc.target);
        cpointer_string!(
            mnt_fstype_cptr,
            mnt_fstype_cstring,
            mount_desc.partition_type.to_string()
        );

        let mount_flags = mount_desc
            .options
            .iter()
            .fold(0, |c, o| c | (*o).clone() as i64);

        log::debug!("mounting {} on {}", mount_desc.source, mount_desc.target);

        let syscall_result = unsafe {
            match (mount_desc.partition_type, mount_desc.data) {
                (FsType::NoneFs, None) => syscall!(
                    SYS_mount,
                    mnt_source_cptr,
                    mnt_target_cptr,
                    0,
                    mount_flags,
                    0
                ),
                (_, None) => syscall!(
                    SYS_mount,
                    mnt_source_cptr,
                    mnt_target_cptr,
                    mnt_fstype_cptr,
                    mount_flags,
                    0
                ),
                (FsType::NoneFs, Some(data)) => {
                    cpointer_string!(mnt_data_cptr, mnt_data_cstring, data);
                    syscall!(
                        SYS_mount,
                        mnt_source_cptr,
                        mnt_target_cptr,
                        0,
                        mount_flags,
                        mnt_data_cptr
                    )
                }
                (_, Some(data)) => {
                    cpointer_string!(mnt_data_cptr, mnt_data_cstring, data);
                    syscall!(
                        SYS_mount,
                        mnt_source_cptr,
                        mnt_target_cptr,
                        mnt_fstype_cptr,
                        mount_flags,
                        mnt_data_cptr
                    )
                }
            }
        };

        syscall_result.map(|_| {}).map_err(|errno| {
            Error::ProviderError(format!("mount, errno: {:?}", Errno::from_i64(errno)))
        })
    }

    pub fn chdir(&self, newdir: String) -> Result<()> {
        cpointer_string!(newdir_cptr, newdir_cstring, newdir.to_string());

        let syscall_result = unsafe { syscall!(SYS_chdir, newdir_cptr) };

        syscall_result.map(|_| {}).map_err(|errno| {
            Error::ProviderError(format!("chdir, errno: {:?}", Errno::from_i64(errno)))
        })
    }
}
